
from vsg.vhdlFile import utils


def classify_selected_name(iToken, lObjects, token):
    iTokenIndex = utils.find_next_token(iToken, lObjects)
    lTokens = lObjects[iTokenIndex].get_value().split('.')
    if lObjects[iTokenIndex + 1].get_value().startswith('"'):
        lTokens[-1] = lObjects.pop(iTokenIndex + 1).get_value()
    lNewTokens = build_selected_name_token_list(lTokens, token)
    replace_item_in_list_with_a_list_at_index(lObjects, lNewTokens, iTokenIndex)
    iNewIndex = iToken + len(lNewTokens)

    return iNewIndex


def build_selected_name_token_list(lTokens, token):
    if is_use_clause_selected_name(token):
        return build_use_clause_selected_name_token_list(lTokens, token)
    return build_context_reference_selected_name_token_list(lTokens, token)


def build_use_clause_selected_name_token_list(lTokens, token):
    lNewTokens = []
    for iThisToken, sToken in enumerate(lTokens):
        classify_use_clause_selected_name_elements(iThisToken, lNewTokens, lTokens, token)
    lNewTokens.pop()
    return lNewTokens


def build_context_reference_selected_name_token_list(lTokens, token):
    lNewTokens = []
    for iThisToken, sToken in enumerate(lTokens):
        classify_context_reference_selected_name_elements(iThisToken, lNewTokens, lTokens, token)
    lNewTokens.pop()
    return lNewTokens


def classify_use_clause_selected_name_elements(iThisToken, lNewTokens, lTokens, token):
    sToken = lTokens[iThisToken]
    if iThisToken == 0:
        lNewTokens.append(token.library_name(sToken))
    elif sToken.lower() == 'all':
        lNewTokens.append(token.all_keyword(sToken))
    elif iThisToken == len(lTokens) - 1:
        lNewTokens.append(token.item_name(sToken))
    else:
        lNewTokens.append(token.package_name(sToken))
    lNewTokens.append(token.dot())


def classify_context_reference_selected_name_elements(iThisToken, lNewTokens, lTokens, token):
    sToken = lTokens[iThisToken]
    if iThisToken == 0 and len(lTokens) > 1:
        lNewTokens.append(token.library_name(sToken))
    else:
        lNewTokens.append(token.context_name(sToken))
    lNewTokens.append(token.dot())


def replace_item_in_list_with_a_list_at_index(lFirstList, lSecondList, iIndex):
    lFirstList.pop(iIndex)
    lSecondList.reverse()
    for oToken in lSecondList:
        lFirstList.insert(iIndex, oToken)


def is_use_clause_selected_name(token):
    if token.__name__ == 'vsg.token.use_clause':
        return True
    return False


def classify_production(production, iToken, lObjects):
    iCurrent = iToken
    iStop = len(lObjects)
    while iCurrent < iStop:
        iPrevious = iCurrent
        iCurrent = production.detect(iCurrent, lObjects)
        if iPrevious == iCurrent:
            break
    return iCurrent




from vsg import parser


class keyword(parser.keyword):
    '''
    unique_id = assertion : keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class condition(parser.condition):
    '''
    unique_id = assertion : condition
    '''

    def __init__(self, sString):
        parser.condition.__init__(self, sString)


class report_keyword(parser.keyword):
    '''
    unique_id = assertion : report_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class report_expression(parser.expression):
    '''
    unique_id = assertion : report_expression
    '''

    def __init__(self, sString):
        parser.expression.__init__(self, sString)


class severity_keyword(parser.keyword):
    '''
    unique_id = assertion : severity_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class severity_expression(parser.expression):
    '''
    unique_id = assertion : severity_expression
    '''

    def __init__(self, sString):
        parser.expression.__init__(self, sString)



#!/usr/bin/env python

import sys
import os
import json
import glob
import yaml
import functools
import multiprocessing

from . import apply_rules
from . import cmd_line_args
from . import config
from . import utils

from . import junit
from . import rule_list
from . import severity
from . import version
from . import vhdlFile

from vsg.report import quality_report


def generate_output_configuration(commandLineArguments, oConfig):
    '''
    Creates a configuration based on parameters passed on the command line.
    It will send the output to a file in JSON format.

    Parameters:

      commandLineArguments: (argparse object)

      configuration: (configuration dictionary)

    Returns:  Nothing
    '''
    if commandLineArguments.output_configuration:
        configuration = oConfig.dConfig
        fExitStatus = 0
        # Create empty file so it can be used to create the rule list
        oVhdlFile = vhdlFile.vhdlFile([''])
        oRules = rule_list.rule_list(oVhdlFile, oConfig.severity_list, commandLineArguments.local_rules)
        oRules.configure(oConfig)
        dOutputConfiguration = {}
        if commandLineArguments.filename:
            dOutputConfiguration['file_list'] = []
            for sFileName in commandLineArguments.filename:
                dOutputConfiguration['file_list'].append(sFileName)
        if commandLineArguments.local_rules:
            dOutputConfiguration['local_rules'] = commandLineArguments.local_rules
        dOutputConfiguration['rule'] = oRules.get_configuration()
        dOutputConfiguration['indent'] = configuration['indent']
        with open(commandLineArguments.output_configuration, 'w') as json_file:
            json.dump(dOutputConfiguration, json_file, sort_keys=True, indent=2)
        sys.exit(fExitStatus)


def display_rule_configuration(commandLineArguments, oConfig):
    '''
    Displays the configuration of a rule passed on the command line.

    Parameters:

      commandLineArguments: (argparse object)

      configuration: (configuration dictionary)

    Returns:  Nothing
    '''
    if commandLineArguments.rule_configuration:
        configuration = oConfig.dConfig
        fExitStatus = 0
        # Create empty file so it can be used to create the rule list
        oVhdlFile = vhdlFile.vhdlFile([''])
        oRules = rule_list.rule_list(oVhdlFile, oConfig.severity_list, commandLineArguments.local_rules)
        oRules.configure(oConfig)
        dOutputConfiguration = {}
        if commandLineArguments.local_rules:
            dOutputConfiguration['local_rules'] = commandLineArguments.local_rules
        dFullConfiguration = oRules.get_configuration()
        if commandLineArguments.rule_configuration in dFullConfiguration:
            dOutputConfiguration['rule'] = {}
            dOutputConfiguration['rule'][commandLineArguments.rule_configuration] = dFullConfiguration[commandLineArguments.rule_configuration]
            # Format the data for displaying
            print(json.dumps(dOutputConfiguration, indent=2))
        else:
            print('ERROR: rule ' + commandLineArguments.rule_configuration + ' was not found.')
            fExitStatus = 1
        sys.exit(fExitStatus)


def validate_files_exist_to_analyze(sName):
    if sName is None:
        print('ERROR: No file defined by the -f command line option or filename given in configuration file.')
        sys.exit(1)


def main():
    '''Main routine of the VHDL Style Guide (VSG) program.'''

    fExitStatus = 0

    commandLineArguments = cmd_line_args.parse_command_line_arguments()

    version.print_version(commandLineArguments)

    oConfig = config.New(commandLineArguments)

    # Add local rule path to system path so the rules can be loaded
    if commandLineArguments.local_rules:
        sys.path.append(os.path.abspath(commandLineArguments.local_rules))

    if commandLineArguments.junit:
        oJunitFile = junit.xmlfile(commandLineArguments.junit)
        oJunitTestsuite = junit.testsuite('vhdl-style-guide', str(0))

    generate_output_configuration(commandLineArguments, oConfig)

    display_rule_configuration(commandLineArguments, oConfig)

    validate_files_exist_to_analyze(commandLineArguments.filename)

    dJson = create_json_dictionary()

    f = functools.partial(apply_rules.apply_rules, commandLineArguments, oConfig)
    # It's easier to debug when not using multiprocessing.Pool()
    lReturn = []
    if commandLineArguments.jobs == 1:
        for iIndex, sFileName in enumerate(commandLineArguments.filename):
            fStatus, testCase, dJsonEntry, sOutputStd, sOutputErr = f((iIndex, sFileName))
            lReturn.append((fStatus, testCase, dJsonEntry))
            if sOutputStd:
                print(sOutputStd)
            if sOutputErr:
                print(sOutputErr, file=sys.stderr)

    else:
        with multiprocessing.Pool(commandLineArguments.jobs) as pool:
            for tResult in pool.imap(f, enumerate(commandLineArguments.filename)):
                fStatus, testCase, dJsonEntry, sOutputStd, sOutputErr = tResult
                lReturn.append((fStatus, testCase, dJsonEntry))
                if sOutputStd:
                    print(sOutputStd)
                if sOutputErr:
                    print(sOutputErr, file=sys.stderr)

    for tValue in lReturn:
        fStatus, testCase, dJsonEntry = tValue
        fExitStatus = fExitStatus or fStatus

        if commandLineArguments.junit:
            oJunitTestsuite.add_testcase(testCase)

        if commandLineArguments.json or commandLineArguments.quality_report:
            dJson['files'].append(dJsonEntry)

    if commandLineArguments.junit:
        oJunitFile.add_testsuite(oJunitTestsuite)
        utils.write_junit_xml_file(oJunitFile)

    if commandLineArguments.json:
        with open(commandLineArguments.json, 'w') as oFile:
            oFile.write(json.dumps(dJson, indent=2))

    if commandLineArguments.quality_report:
        quality_report.write(commandLineArguments, dJson)

    sys.exit(fExitStatus)


def create_json_dictionary():
    '''
    Returns a dictionary to be used when the --json option is chosen.
    '''
    dJson = {}
    dJson['files'] = []
    return dJson




from vsg.token import port_clause as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import port_list


def detect(iToken, lObjects):
    '''
    port_clause ::=
        port ( port_list ) ;
    '''

    if utils.are_next_consecutive_tokens(['port', '('], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('port', token.port_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)

    iCurrent = port_list.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import blank_line_below_line_ending_with_token

from vsg import token

lTokens = []
lTokens.append(token.subprogram_body.begin_keyword)


class rule_203(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **begin** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       function overflow (a: integer) return integer is
       begin
         a <= b;

    **Fix**

    .. code-block:: vhdl

       function overflow (a: integer) return integer is
       begin

         a <= b;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'subprogram_body', '203', lTokens)




from vsg import parser

from vsg.token import constant_declaration

from vsg.rules import token_indent_unless_between_tokens

lTokens = []
lTokens.append(parser.comment)

lUnless = []
lUnless.append([constant_declaration.constant_keyword, constant_declaration.semicolon])


class rule_010(token_indent_unless_between_tokens):
    '''
    This rule checks the indent lines starting with comments.

    **Violation**

    .. code-block:: vhdl

           -- Libraries
       libary ieee;

        -- Define architecture
       architecture rtl of fifo is

       -- Define signals
         signal wr_en : std_logic;
         signal rd_en : std_Logic;

       begin

    **Fix**

    .. code-block:: vhdl

       -- Libraries
       libary ieee;

       -- Define architecture
       architecture rtl of fifo is

         -- Define signals
         signal wr_en : std_logic;
         signal rd_en : std_Logic;

       begin
    '''

    def __init__(self):
        token_indent_unless_between_tokens.__init__(self, 'comment', '010', lTokens, lUnless)
        self.phase = 4
        self.subphase = 3




from vsg import proposed_rule


class rule_007(proposed_rule.Rule):
    '''
    This rule checks for code after the semicolon.

    .. NOTE:: This rule has not been implemented yet.

    **Violation**

    .. code-block:: vhdl

       context c1; -- Comments are allowed

       context c1; library ieee; -- This is not allowed

    **Fix**

    .. code-block:: vhdl

       context c1; -- Comments are allowed

       context c1;
         library ieee; -- This is not allowed
    '''

    def __init__(self):
        proposed_rule.Rule.__init__(self, 'context_ref', '007')




from vsg import parser


class generate_label(parser.label):
    '''
    unique_id = case_generate_statement : generate_label
    '''

    def __init__(self, sString):
        parser.label.__init__(self, sString)


class label_colon(parser.label_colon):
    '''
    unique_id = case_generate_statement : label_colon
    '''

    def __init__(self):
        parser.label_colon.__init__(self)


class case_keyword(parser.keyword):
    '''
    unique_id = case_generate_statement : case_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class generate_keyword(parser.keyword):
    '''
    unique_id = case_generate_statement : generate_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_keyword(parser.keyword):
    '''
    unique_id = case_generate_statement : end_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_generate_keyword(parser.keyword):
    '''
    unique_id = case_generate_statement : end_generate_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_generate_label(parser.label):
    '''
    unique_id = case_generate_statement : end_generate_label
    '''

    def __init__(self, sString):
        parser.label.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = case_generate_statement : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.token import entity_aspect as token

from vsg.vhdlFile import utils


def classify(iToken, lObjects):
    '''
    entity_aspect ::=
        **entity** entity_name [ ( architecture_identifier ) ]
      | **configuration** configuration_name
      | **open**
    '''
    iCurrent = iToken

    if utils.is_next_token('open', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('open', token.open_keyword, iCurrent, lObjects)

    elif utils.is_next_token('configuration', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('configuration', token.configuration_keyword, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(token.configuration_name, iCurrent, lObjects)

    elif utils.is_next_token('entity', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('entity', token.entity_keyword, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(token.entity_name, iCurrent, lObjects)

        if utils.is_next_token('(', iCurrent, lObjects):
            iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)
            iCurrent = utils.assign_next_token(token.architecture_identifier, iCurrent, lObjects)
            iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)

    return iCurrent




from vsg import token
from vsg import violation

from vsg.rule_group import structure


class rule_001(structure.Rule):
    '''
    This rule checks for **with** statements.

    **Violation**

    .. code-block:: vhdl

       with buttons select

    **Fix**

    Refactor **with** statement into a process.
    '''

    def __init__(self):
        structure.Rule.__init__(self, name='with', identifier='001')
        self.solution = "Rewrite with as a process"
        self.fixable = False

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_matching([token.concurrent_selected_signal_assignment.with_keyword])

    def _analyze(self, lToi):
        for oToi in lToi:
            self.add_violation(violation.New(oToi.get_line_number(), oToi, self.solution))




from vsg.token import block_configuration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import utils as c_utils

from vsg.vhdlFile.classify import block_specification
from vsg.vhdlFile.classify import use_clause
from vsg.vhdlFile.classify import configuration_item


def detect(iToken, lObjects):
    '''
    block_configuration ::=
        for block_specification
            { use_clause }
            { configuration_item }
        end for ;
    '''

    if utils.is_next_token('for', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('for', token.for_keyword, iToken, lObjects)

    iCurrent = block_specification.classify(iCurrent, lObjects)

    iCurrent = c_utils.classify_production(use_clause, iCurrent, lObjects)
    iCurrent = c_utils.classify_production(configuration_item, iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('for', token.end_for_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.unspecified, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.case_generate_statement.end_keyword, token.case_generate_statement.end_generate_keyword])
lTokens.append([token.for_generate_statement.end_keyword, token.for_generate_statement.end_generate_keyword])
lTokens.append([token.if_generate_statement.end_keyword, token.if_generate_statement.end_generate_keyword])


class rule_008(Rule):
    '''
    This rule checks for a single space after the **end** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end   generate ram_array;

    **Fix**

    .. code-block:: vhdl

       end generate ram_array;
    '''
    def __init__(self):
        Rule.__init__(self, 'generate', '008', lTokens)




from vsg.rules.whitespace_after_token import Rule

from vsg import token

lTokens = []
lTokens.append(token.constant_declaration.assignment_operator)


class rule_100(Rule):
    '''
    This rule checks for a single space after the := assignment in constant declarations.
    Having a space makes it clearer where the assignment occurs on the line.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       constant size : integer :=1;
       constant width : t_type :=(

    **Fix**

    .. code-block:: vhdl

       constant size : integer := 1;
       constant width : t_type := (
    '''

    def __init__(self):
        Rule.__init__(self, 'constant', '100', lTokens)




import glob
import os
import sys
import yaml

from . import junit
from . import severity
from . import utils


def read_predefined_style(sStyleName):
    '''
    Reads a predefined style file.

    Parameters :

      sStyleName : (string)

    Returns : (dictionary)
    '''
    dReturn = {}
    if sStyleName is not None:
        sFileName = os.path.join(os.path.dirname(__file__), 'styles', sStyleName + '.yaml')
        dReturn = open_configuration_file(sFileName)
    return dReturn


def open_configuration_file(sFileName, sJUnitFileName=None):
    '''Attempts to open a configuration file and read it's contents.'''
    try:
        with open(sFileName) as yaml_file:
            tempConfiguration = yaml.full_load(yaml_file)
    except OSError as e:
        print(f'ERROR: encountered {e.__class__.__name__}, {e.args[1]} while opening configuration file: ' + sFileName)
        write_invalid_configuration_junit_file(sFileName, sJUnitFileName)
        sys.exit(1)
    except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
        print('ERROR: Invalid configuration file: ' + sFileName)
        print(e)
        write_invalid_configuration_junit_file(sFileName, sJUnitFileName)
        sys.exit(1)
    return tempConfiguration


def validate_file_exists(sFilename, sConfigName):
    '''Validates a file exist while using the glob function to expand filenames.'''
    if isinstance(sFilename, dict):
        sExpandedFilename = list(sFilename.keys())[0]
    else:
        sExpandedFilename = sFilename
#    print(sExpandedFilename)
    lFileNames = glob.glob(utils.expand_filename(sExpandedFilename), recursive=True)
#    print(lFileNames)
    if len(lFileNames) == 0:
        print('ERROR: Could not find file ' + sFilename + ' in configuration file ' + sConfigName)
        sys.exit(1)


def read_configuration_files(dStyle, commandLineArguments):
    if not commandLineArguments.configuration:
        return dStyle

    dConfiguration = dStyle
    for sConfigFilename in commandLineArguments.configuration:
        tempConfiguration = open_configuration_file(sConfigFilename, commandLineArguments.junit)

        dConfiguration = process_config_file(dConfiguration, tempConfiguration, sConfigFilename)

    return dConfiguration


def process_config_file(dConfiguration, tempConfiguration, sConfigFilename):
    dReturn = dConfiguration
    for sKey in tempConfiguration.keys():
        if sKey == 'file_list':
            dReturn = process_file_list_key(dReturn, tempConfiguration, sKey, sConfigFilename)

        elif sKey == 'rule':
            for sRule in tempConfiguration[sKey]:
                try:
                    dReturn[sKey][sRule] = tempConfiguration[sKey][sRule]
                except KeyError:
                    dReturn[sKey] = {}
                    dReturn[sKey][sRule] = tempConfiguration[sKey][sRule]
        else:
            dReturn[sKey] = tempConfiguration[sKey]
    return dReturn


def process_file_list_key(dConfig, tempConfiguration, sKey, sConfigFilename):
    dReturn = dConfig
    if 'file_list' not in dConfig:
        dReturn['file_list'] = []
    for iIndex, sFilename in enumerate(tempConfiguration['file_list']):
        validate_file_exists(sFilename, sConfigFilename)
        try:
            for sGlobbedFilename in glob.glob(utils.expand_filename(sFilename), recursive=True):
                dReturn['file_list'].append(sGlobbedFilename)
        except TypeError:
            sKey = list(sFilename.keys())[0]
            for sGlobbedFilename in glob.glob(utils.expand_filename(sKey), recursive=True):
                dTemp = {}
                dTemp[sGlobbedFilename] = {}
                dTemp[sGlobbedFilename].update(tempConfiguration['file_list'][iIndex][sKey])
                dReturn['file_list'].append(dTemp)
    return dReturn


def write_invalid_configuration_junit_file(sFileName, sJUnitFileName):
    if sJUnitFileName:
        oJunitFile = junit.xmlfile(sJUnitFileName)
        oJunitTestsuite = junit.testsuite('vhdl-style-guide', str(0))
        oJunitTestcase = junit.testcase(sFileName, str(0), 'failure')
        oFailure = junit.failure('Failure')
        oFailure.add_text('Invalid JSON format.  Review configuration for errors.')
        oJunitTestcase.add_failure(oFailure)
        oJunitTestsuite.add_testcase(oJunitTestcase)
        oJunitFile.add_testsuite(oJunitTestsuite)
        utils.write_junit_xml_file(oJunitFile)


def add_debug_to_configuration(oCLA, dConfiguration):
    '''
    Adds debug values to the configuration dictionary for later use.

    Parameters:

      oCLA: (command line argument object)

      dConfiguration: (dictionary)

    Returns:  Nothing
    '''
    try:
        dConfiguration['debug'] = oCLA.debug
    except TypeError:
        dConfiguration = {}
        dConfiguration['debug'] = oCLA.debug
    return dConfiguration


def read_indent_configuration(dConfiguration):
    '''
    Reads the default indent dictionary and merges any changes from a user configuration.

    Parameters:

       dConfiguration: (dictionary)

    Returns:  (dictionary)
    '''

    sFileName = os.path.join(os.path.dirname(__file__), 'vhdlFile', 'indent', 'indent_config.yaml')

    dReturn = open_configuration_file(sFileName)

    ### This merges an indent configuration into the base indent dictionary
    if 'indent' in list(dConfiguration.keys()):
        dGroups = dConfiguration['indent']['tokens']
        for sGroup in list(dGroups.keys()):
            for sToken in list(dGroups[sGroup].keys()):
                for sParameter in list(dGroups[sGroup][sToken].keys()):
                    dReturn['indent']['tokens'][sGroup][sToken][sParameter] = dGroups[sGroup][sToken][sParameter]

    dConfiguration['indent'] = dReturn['indent']

    return dReturn


def update_command_line_arguments(commandLineArguments, configuration):

    if 'skip_phase' in configuration:
        commandLineArguments.skip_phase = configuration['skip_phase']
    else:
        commandLineArguments.skip_phase = []

    if not configuration:
        return

    if 'file_list' in configuration:
        for sFilename in configuration['file_list']:
            if isinstance(sFilename, dict):
                sFilename = list(sFilename.keys())[0]
            try:
                lFileNames = glob.glob(utils.expand_filename(sFilename), recursive=True)
                for sFileName in lFileNames:
                    if sFileName not in commandLineArguments.filename:
                        commandLineArguments.filename.append(sFileName)
            except:
                commandLineArguments.filename = glob.glob(utils.expand_filename(sFilename), recursive=True)

    if 'local_rules' in configuration:
        commandLineArguments.local_rules = utils.expand_filename(configuration['local_rules'])


def New(commandLineArguments):
    oReturn = config()

    dStyle = read_predefined_style(commandLineArguments.style)
    dConfig = read_configuration_files(dStyle, commandLineArguments)

    oReturn.severity_list = severity.create_list(dConfig)

    update_command_line_arguments(commandLineArguments, dConfig)

    dConfig = add_debug_to_configuration(commandLineArguments, dConfig)

    oReturn.dConfig = dConfig

    dIndent = read_indent_configuration(dConfig)
    oReturn.dIndent = dIndent

    if commandLineArguments.fix_only:
        oReturn.dFixOnly = open_configuration_file(commandLineArguments.fix_only)
    else:
        oReturn.dFixOnly = None

    return oReturn


class config():

    def __init__(self):
        dIndent = None
        dConfig = None
        dFixOnly = None
        severity_list = None




from vsg.token import interface_subprogram_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import interface_subprogram_specification
from vsg.vhdlFile.classify import interface_subprogram_default


def detect(iToken, lObjects):
    '''
    interface_subprogram_declaration ::=
        interface_subprogram_specification [ is interface_subprogram_default ]
    '''
    iCurrent = utils.find_next_token(iToken, lObjects)
    iLast = iCurrent
    iCurrent = interface_subprogram_specification.detect(iCurrent, lObjects)
    if iLast != iCurrent:
        iCurrent = utils.find_next_token(iToken, lObjects)
        if utils.object_value_is(lObjects, iCurrent, 'is'):
            return classify(iCurrent, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_if('is', token.is_keyword, iToken, lObjects)
    iCurrent = interface_subprogram_default.classify(iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_indent

from vsg.token import architecture_body as token


class rule_007(token_indent):
    '''
    This rule checks for spaces before the **begin** keyword.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is
         begin

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin
    '''

    def __init__(self):
        token_indent.__init__(self, 'architecture', '007', [token.begin_keyword])




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class rule_012(alignment.Rule):
    '''
    This rule checks multiple signal declarations on a single line are column aligned.

    .. NOTE::
        This rule will only cover two signals on a single line.

    **Violation**

    .. code-block:: vhdl

       signal wr_en, wr_en_f             : std_logic;
       signal rd_en_f, rd_en             : std_logic;
       signal chip_select, chip_select_f : t_user_defined_type;

    **Fix**

    .. code-block:: vhdl

       signal wr_en,       wr_en_f       : std_logic;
       signal rd_en_f,     rd_en         : std_logic;
       signal chip_select, chip_select_f : t_user_defined_type;
    '''

    def __init__(self):
        alignment.Rule.__init__(self, 'signal', '012')
        self.subphase = 2
        self.left_token = token.architecture_body.is_keyword
        self.right_token = token.architecture_body.begin_keyword
        self.lUnless = []
        self.lUnless.append([token.subprogram_body.is_keyword,token.subprogram_body.begin_keyword])
        ## attributes below are configurable by the user

        self.compact_alignment = True
        self.configuration.append('compact_alignment')

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.left_token, self.right_token)

    def _analyze(self, lToi):
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            iColumn = 0
            bSignalFound = False
            bSkip = False
            dAnalysis = {}
            dTemp = {}
            for iToken, oToken in enumerate(lTokens):

               iLine = utils.increment_line_number(iLine, oToken)

               if isinstance(oToken, parser.carriage_return):
                   iColumn = 0
               else:
                   iColumn += len(oToken.get_value())

               bSkip = check_for_exclusions(bSkip, oToken, self.lUnless)
               if bSkip:
                   continue

               bSignalFound = check_for_signal_declaration(bSignalFound, oToken)
               if not bSignalFound:
                   iComma = 0
                   continue

               if isinstance(oToken, token.signal_declaration.colon):
                   bSignalFound = False
                   if iComma == 1:
                       dAnalysis[dTemp['line_number']] = dTemp
                   continue

               if isinstance(oToken, parser.comma):
                   iComma += 1
                   if iComma == 2:
                       bSignalFound = False
                       continue

                   dTemp = {}
                   dTemp['comma_column'] = iColumn
                   dTemp['comma_index'] = iToken
                   dTemp['line_number'] = iLine
                   if utils.are_next_consecutive_token_types([parser.whitespace, token.signal_declaration.identifier], iToken + 1, lTokens):
                       dTemp['identifier_column'] = iColumn + len(lTokens[iToken + 1].get_value())
                       dTemp['token_index'] = iToken + 2
                       dTemp['token_value'] = lTokens[iToken + 2].get_value()
                   elif utils.are_next_consecutive_token_types([token.signal_declaration.identifier], iToken + 1, lTokens):
                       dTemp['identifier_column'] = iColumn + 1
                       dTemp['token_index'] = iToken + 1
                       dTemp['token_value'] = lTokens[iToken + 1].get_value()
                   else:
                       bSignalFound = False

            add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

            for iKey in list(dAnalysis.keys()):
                if dAnalysis[iKey]['adjust'] != 0:
                    oLineTokens = oToi.extract_tokens(dAnalysis[iKey]['comma_index'], dAnalysis[iKey]['token_index'])
                    sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                    oViolation = violation.New(dAnalysis[iKey]['line_number'], oLineTokens, sSolution)
                    oViolation.set_action(dAnalysis[iKey])
                    self.add_violation(oViolation)

            dAnalysis = {}

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()

        if len(lTokens) == 2:
            rules_utils.insert_whitespace(lTokens, 1)
        else:
            iLen = len(lTokens[1].get_value()) + dAction['adjust']
            lTokens[1].set_value(' '*iLen)

        oViolation.set_tokens(lTokens)


def add_adjustments_to_dAnalysis(dAnalysis, compact_alignment):
    iMaxLeftColumn = 0
    iMinLeftColumn = 9999999999999999
    iMaxTokenColumn = 0
    iMinTokenColumn = 9999999999999999

    for iKey in list(dAnalysis.keys()):
        iMaxLeftColumn = max(iMaxLeftColumn, dAnalysis[iKey]['comma_column'])
        iMinLeftColumn = min(iMinLeftColumn, dAnalysis[iKey]['comma_column'])
        iMaxTokenColumn = max(iMaxTokenColumn, dAnalysis[iKey]['identifier_column'])
        iMinTokenColumn = min(iMinTokenColumn, dAnalysis[iKey]['identifier_column'])

    if compact_alignment:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxLeftColumn - dAnalysis[iKey]['identifier_column'] + 1
    else:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxTokenColumn - dAnalysis[iKey]['identifier_column'] + 1


def check_for_exclusions(bSkip, oToken, lUnless):
    for lTokenPairs in lUnless:
        if isinstance(oToken, lTokenPairs[0]):
            return True
        if isinstance(oToken, lTokenPairs[1]):
            return False
    return bSkip


def check_for_signal_declaration(bSignalFound, oToken):
    if isinstance(oToken, token.signal_declaration.signal_keyword):
        return True
    if isinstance(oToken, token.signal_declaration.semicolon):
        return False
    return bSignalFound




from vsg.vhdlFile.classify import association_list


def classify(iTokent, lObjects):
    '''
    actual_parameter_part ::=
        *parameter*_association_list
    '''

    return association_list.classify(iTokent, lObjects)




from vsg.rules import move_token_to_the_right_of_several_possible_tokens_if_it_exists_between_tokens as Rule

from vsg import token

lAnchorTokens = []
lAnchorTokens.append(token.loop_statement.end_loop_keyword)
lAnchorTokens.append(token.loop_statement.end_loop_label)

oToken = token.loop_statement.semicolon

oStartToken = token.loop_statement.end_loop_keyword
oEndToken = token.loop_statement.semicolon


class rule_004(Rule):
    '''
    This rule checks the semicolon is on the same line as the **end loop** keyword.

    **Violation**

    .. code-block:: vhdl

       end loop
       ;

       end loop LOOP_LABEL
       ;

    **Fix**

    .. code-block:: vhdl

       end loop;

       end loop LOOP_LABEL;
    '''

    def __init__(self):
        Rule.__init__(self, 'loop_statement', '004', oToken, lAnchorTokens, oStartToken, oEndToken)
        self.subphase = 3




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.function_specification.close_parenthesis)


class rule_300(token_indent):
    '''
    This rule checks the indent of the closing parenthesis if it is on its own line.

    **Violation**

    .. code-block:: vhdl

       function func_1 (a : integer; b : integer;
         c : unsigned(3 downto 0);
         d : std_logic_vector(7 downto 0);
         e : std_logic
         ) return integer is


    **Fix**

    .. code-block:: vhdl

       function func_1 (a : integer; b : integer;
         c : unsigned(3 downto 0);
         d : std_logic_vector(7 downto 0);
         e : std_logic
       ) return integer is
    '''

    def __init__(self):
        token_indent.__init__(self, 'function', '300', lTokens)




from vsg import token

from vsg.rules import token_case_with_prefix_suffix

lTokens = []
lTokens.append(token.case_generate_statement.generate_label)
lTokens.append(token.for_generate_statement.generate_label)
lTokens.append(token.if_generate_statement.generate_label)


class rule_005(token_case_with_prefix_suffix):
    '''
    This rule checks the generate label has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       RAM_ARRAY: for i in 0 to 7 generate

    **Fix**

    .. code-block:: vhdl

       ram_array: for i in 0 to 7 generate
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'generate', '005', lTokens)
        self.groups.append('case::label')




from vsg.rules import insert_token_right_of_token_if_it_does_not_exist_before_token

from vsg.token import component_declaration as token


class rule_021(insert_token_right_of_token_if_it_does_not_exist_before_token):
    '''
    This rule inserts the optional **is** keyword if it does not exist.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       component my_component

       end my_component;

    **Fix**

    .. code-block:: vhdl

       component my_component is

       end my_component;
    '''
    def __init__(self):
        insert_token_right_of_token_if_it_does_not_exist_before_token.__init__(self, 'component', '021', token.is_keyword('is'), token.identifier, token.semicolon)
        self.solution = '*is* keyword.'
        self.groups.append('structure::optional')




from vsg.rules import token_case_in_range_bounded_by_tokens

from vsg import token

lTokens = []
lTokens.append(token.mode.in_keyword)
lTokens.append(token.mode.out_keyword)
lTokens.append(token.mode.inout_keyword)
lTokens.append(token.mode.buffer_keyword)
lTokens.append(token.mode.linkage_keyword)

oStart = token.port_clause.open_parenthesis
oEnd = token.port_clause.close_parenthesis


class rule_019(token_case_in_range_bounded_by_tokens):
    '''
    This rule checks the port direction has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         WR_EN    : IN    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : OUT   std_logic;
         DATA     : INOUT std_logic
       );

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic
       );
    '''

    def __init__(self):
        token_case_in_range_bounded_by_tokens.__init__(self, 'port', '019', lTokens, oStart, oEnd)
        self.groups.append('case::keyword')




from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.block_statement.end_block_label)


class rule_506(token_case_with_prefix_suffix):
    '''
    This rule checks the label has proper case on the end block declaration.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end block BLOCK_LABEL;

    **Fix**

    .. code-block:: vhdl

       end block block_label;
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'block', '506', lTokens)
        self.groups.append('case::label')




from vsg.rules import multiline_alignment_between_tokens

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.simple_waveform_assignment.assignment, token.simple_waveform_assignment.semicolon])
lTokenPairs.append([token.simple_force_assignment.assignment, token.simple_force_assignment.semicolon])
lTokenPairs.append([token.conditional_waveform_assignment.assignment, token.conditional_waveform_assignment.semicolon])
lTokenPairs.append([token.conditional_force_assignment.assignment, token.conditional_force_assignment.semicolon])


class rule_004(multiline_alignment_between_tokens):
    '''
    This rule checks the alignment of multiline sequential statements.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       overflow <= wr_en and
         rd_en;

    **Fix**

    .. code-block:: vhdl

       overflow <= wr_en and
                   rd_en;
    '''

    def __init__(self):
        multiline_alignment_between_tokens.__init__(self, 'sequential', '004', lTokenPairs)
        self.phase = 5
        self.subphase = 2




from vsg.vhdlFile.classify import subprogram_declaration
from vsg.vhdlFile.classify import subprogram_instantiation_declaration
from vsg.vhdlFile.classify import package_declaration
from vsg.vhdlFile.classify import package_instantiation_declaration
from vsg.vhdlFile.classify import type_declaration
from vsg.vhdlFile.classify import subtype_declaration
from vsg.vhdlFile.classify import constant_declaration
from vsg.vhdlFile.classify import signal_declaration
from vsg.vhdlFile.classify import variable_declaration
from vsg.vhdlFile.classify import file_declaration
from vsg.vhdlFile.classify import alias_declaration
from vsg.vhdlFile.classify import component_declaration
from vsg.vhdlFile.classify import attribute_declaration
from vsg.vhdlFile.classify import attribute_specification

from vsg.vhdlFile.classify import use_clause


def detect(iToken, lObjects):
    '''
    package_declarative_item ::=
        subprogram_declaration
      | subprogram_instantiation_declaration
      | package_declaration
      | package_instantiation_declaration
      | type_declaration
      | subtype_declaration
      | constant_declaration
      | signal_declaration
      | variable_declaration
      | file_declaration
      | alias_declaration
      | component_declaration
      | attribute_declaration
      | attribute_specification
      | disconnection_specification
      | use_clause
      | group_template_declaration
      | group_declaration
      | PSL_Property_Declaration
      | PSL_Sequence_Declaration
    '''

    iReturn = subprogram_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = subprogram_instantiation_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = package_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = package_instantiation_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = type_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = subtype_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = constant_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = signal_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = variable_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = file_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = alias_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = component_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = attribute_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = attribute_specification.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = use_clause.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    return iToken




from vsg.rules import split_line_at_token

from vsg import token

lTokens = []
lTokens.append(token.context_declaration.end_keyword)


class rule_008(split_line_at_token):
    '''
    This rule checks the **end** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       context c1 is library ieee; end context c1;

       context c1 is library ieee; end;

    **Fix**

    .. code-block:: vhdl

       context c1 is library ieee;
       end context c1;

       context c1 is library ieee;
       end;
    '''

    def __init__(self):
        split_line_at_token.__init__(self, 'context', '008', lTokens)
        self.solution = 'Move *end* keyword and code after end to the next line'




from vsg import parser


class for_keyword(parser.keyword):
    '''
    unique_id = timeout_clause : for_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg.vhdlFile.classify import block_configuration
from vsg.vhdlFile.classify import component_configuration


def detect(iToken, lObjects):
    return classify(iToken, lObjects)


def classify(iToken, lObjects):
    '''
    configuration_item ::=
        block_configuration
      | component_configuration
    '''

    iCurrent = component_configuration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = block_configuration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken




from vsg import token

from vsg.rules import token_case

lTokens = []
lTokens.append(token.generate_statement_body.end_keyword)


class rule_501(token_case):
    '''
    This rule checks the **end** keyword has the proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       for condition generate
         begin
         END
       end generate;

    **Fix**

    .. code-block:: vhdl

       for condition generate
         begin
         end
       end generate;
    '''

    def __init__(self):
        token_case.__init__(self, 'generate', '501', lTokens)
        self.groups.append('case::keyword')




from vsg.token import block_header as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import generic_clause
from vsg.vhdlFile.classify import generic_map_aspect
from vsg.vhdlFile.classify import port_clause
from vsg.vhdlFile.classify import port_map_aspect


def detect(iToken, lObjects):
    '''
    block_header ::=
        [ generic_clause
        [ generic_map_aspect ; ] ]
        [ port_clause
        [ port_map_aspect ; ] ]
    '''

    iCurrent = generic_clause.detect(iToken, lObjects)

    iLast = iCurrent
    iCurrent = generic_map_aspect.detect(iCurrent, lObjects)
    if iLast != iCurrent:
        iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    iCurrent = port_clause.detect(iCurrent, lObjects)

    iLast = iCurrent
    iCurrent = port_map_aspect.detect(iCurrent, lObjects)
    if iLast != iCurrent:
        iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import insert_token_right_of_possible_tokens_if_it_does_not_exist_before_token

from vsg.token import process_statement as token

oInsertToken = token.is_keyword('is')

lRightTokens = []
lRightTokens.append(token.process_keyword)
lRightTokens.append(token.close_parenthesis)

oBeforeToken = token.begin_keyword


class rule_012(insert_token_right_of_possible_tokens_if_it_does_not_exist_before_token):
    '''
    This rule checks for the existence of the **is** keyword.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       proc_a : process
       begin
       end process;

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        )
       begin
       end process;


    **Fix**

    .. code-block:: vhdl

       proc_a : process is
       begin
       end process;


       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin
       end process;
    '''

    def __init__(self):
        insert_token_right_of_possible_tokens_if_it_does_not_exist_before_token.__init__(self, 'process', '012', oInsertToken, lRightTokens, oBeforeToken)
        self.solution = '*is* keyword'
        self.groups.append('structure::optional')




from vsg import parser

from vsg.vhdlFile.extract import tokens


def get_tokens_bounded_by_token_when_between_tokens(oLeft, oRight, oStart, oEnd, lAllTokens, oTokenMap, include_trailing_whitespace=False):

    lMyPairs = get_token_pairs([oLeft, oRight], [oStart, oEnd], oTokenMap)

    lReturn = []

    for lPair in lMyPairs:
        iLeft = lPair[0]
        iRight = lPair[1]
        iLine = oTokenMap.get_line_number_of_index(iLeft)
        if include_trailing_whitespace:
            if oTokenMap.is_token_at_index(iRight + 1, parser.whitespace):
                lReturn.append(tokens.New(iLeft, iLine, lAllTokens[iLeft:iRight + 2]))
            else:
                lReturn.append(tokens.New(iLeft, iLine, lAllTokens[iLeft:iRight + 1]))
        else:
            lReturn.append(tokens.New(iLeft, iLine, lAllTokens[iLeft:iRight + 1]))

    return lReturn


def get_token_pairs(lInnerPair, lOuterPair, oTokenMap):
    lLeft, lRight = oTokenMap.get_token_pair_indexes(lInnerPair[0], lInnerPair[1])
    lStart, lEnd = oTokenMap.get_token_pair_indexes(lOuterPair[0], lOuterPair[1])

    lPairs = keep_pairs_that_are_within_pairs(lLeft, lRight, lStart, lEnd)

    lReturn = remove_duplicate_pairs(lPairs)
    return lReturn


def keep_pairs_that_are_within_pairs(lLeft, lRight, lStart, lEnd):
    lPairs = []
    for iStart, iEnd in zip(lStart, lEnd):
        find_inner_pairs_inside_outer_pairs([iStart, iEnd], lLeft, lRight, lPairs)
    return lPairs


def find_inner_pairs_inside_outer_pairs(lOuterPair, lLeft, lRight, lPairs):
    for iLeft, iRight in zip(lLeft, lRight):
        if lOuterPair[0] < iLeft and iRight < lOuterPair[1]:
            lPairs.append([iLeft, iRight])


def remove_duplicate_pairs(lPairs):
    lReturn = []
    for lPair in lPairs:
        if lPair not in lReturn:
            lReturn.append(lPair)
    return lReturn




from vsg import parser


class comma(parser.comma):
    '''
    unique_id = identifier_list : comma
    '''

    def __init__(self, sString=','):
        parser.comma.__init__(self)


class identifier(parser.identifier):
    '''
    unique_id = identifier_list : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)




from vsg import deprecated_rule


class rule_016(deprecated_rule.Rule):
    '''
    This rule was deprecated and replaced with rule:

    * `signal_017 <signal_rules.html#signal-017>`_
    '''
    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'signal', '016')
        self.message.append('Rule ' + self.unique_id + ' has been replaced with the following rule:')
        self.message.append('  signal_017')




from vsg import parser

from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_tokens_at_beginning_of_line_matching_between_tokens(lTokens, oStart, oEnd, bInclusive, lAllTokens, oTokenMap):

    lIndexes = utils.get_indexes_of_token_list(lTokens, oTokenMap)

    lStart, lEnd = oTokenMap.get_token_pair_indexes(oStart, oEnd)

    lReturn = []
    for iIndex in lIndexes:
        if utils.is_index_between_indexes(iIndex, lStart, lEnd, bInclusive=bInclusive):
            if oTokenMap.is_token_at_index(parser.carriage_return, iIndex - 1):
                iLine = oTokenMap.get_line_number_of_index(iIndex)
                lReturn.append(tokens.New(iIndex, iLine, [lAllTokens[iIndex]]))
            elif oTokenMap.is_token_at_index(parser.carriage_return, iIndex - 2) and utils.is_token_at_index_whitespace(oTokenMap, iIndex - 1):
                iLine = oTokenMap.get_line_number_of_index(iIndex)
                lReturn.append(tokens.New(iIndex - 1, iLine, lAllTokens[iIndex - 1: iIndex + 1]))

    return lReturn




from vsg import parser
from vsg import token

from vsg.rules import consistent_token_case

lTokens = []
lTokens.append(token.constant_declaration.identifier)

lIgnore = []
lIgnore.append(token.interface_signal_declaration.identifier)
lIgnore.append(token.interface_unknown_declaration.identifier)
lIgnore.append(token.interface_constant_declaration.identifier)
lIgnore.append(token.interface_variable_declaration.identifier)
lIgnore.append(token.association_element.formal_part)
lIgnore.append(token.entity_declaration.identifier)
lIgnore.append(token.entity_declaration.entity_simple_name)
lIgnore.append(token.architecture_body.entity_name)
lIgnore.append(parser.whitespace)
lIgnore.append(parser.carriage_return)
lIgnore.append(parser.blank_line)


class rule_013(consistent_token_case):
    '''
    This rule checks for consistent capitalization of constant names.

    **Violation**

    .. code-block:: vhdl

       architecture RTL of ENTITY1 is

         constant c_size  : integer := 5;
         constant c_ones  : std_logic_vector(c_size - 1 downto 0) := (others => '1');
         constant c_zeros : std_logic_vector(c_size - 1 downto 0) := (others => '0');

         signal data : std_logic_vector(c_size - 1 downto 0);

       begin

         data <= C_ONES;

         PROC_NAME : process () is
         begin

           data <= C_ones;

           if (sig2 = '0') then
             data <= c_Zeros;
           end if;

         end process PROC_NAME;

       end architecture RTL;

    **Fix**

    .. code-block:: vhdl

       architecture RTL of ENTITY1 is

         constant c_size  : integer := 5;
         constant c_ones  : std_logic_vector(c_size - 1 downto 0) := (others => '1');
         constant c_zeros : std_logic_vector(c_size - 1 downto 0) := (others => '0');

         signal data : std_logic_vector(c_size - 1 downto 0);

       begin

         data <= c_ones;

         PROC_NAME : process () is
         begin

           data <= c_ones;

           if (sig2 = '0') then
             data <= c_zeros;
           end if;

         end process PROC_NAME;

       end architecture RTL;
    '''

    def __init__(self):
        consistent_token_case.__init__(self, 'constant', '013', lTokens, lIgnore)




from vsg import parser


class port_keyword(parser.keyword):
    '''
    unique_id = port_clause : port_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class open_parenthesis(parser.open_parenthesis):
    '''
    unique_id = port_clause : open_parenthesis
    '''

    def __init__(self, sString='('):
        parser.open_parenthesis.__init__(self)


class close_parenthesis(parser.close_parenthesis):
    '''
    unique_id = port_clause : close_parenthesis
    '''

    def __init__(self, sString=')'):
        parser.close_parenthesis.__init__(self)


class semicolon(parser.semicolon):
    '''
    unique_id = port_clause : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg import proposed_rule


class rule_008(proposed_rule.Rule):
    '''
    This rule checks the context selected name is on the same line as the **context** keyword.

    .. NOTE:: This rule has not been implemented yet.

    **Violation**

    .. code-block:: vhdl

       context
       c1
       ;

    **Fix**

    .. code-block:: vhdl

       context c1

       ;
    '''

    def __init__(self):
        proposed_rule.Rule.__init__(self, 'context_ref', '008')




from vsg import parser


class package_keyword(parser.keyword):
    '''
    unique_id = interface_package_declaration : package_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class is_keyword(parser.keyword):
    '''
    unique_id = interface_package_declaration : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class new_keyword(parser.keyword):
    '''
    unique_id = interface_package_declaration : new_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class uninstantiated_package_name(parser.name):
    '''
    unique_id = interface_package_declaration : uninstantiated_package_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)




from vsg import parser
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import whitespace


class rule_008(whitespace.Rule):
    '''
    This rule checks for spaces after the **std_logic_vector** keyword.

    **Violation**

    .. code-block:: vhdl

       signal data    : std_logic_vector (7 downto 0);
       signal counter : std_logic_vector    (7 downto 0);

    **Fix**

    .. code-block:: vhdl

       signal data    : std_logic_vector(7 downto 0);
       signal counter : std_logic_vector(7 downto 0);
    '''

    def __init__(self):
        whitespace.Rule.__init__(self, 'whitespace', '008')
        self.solution = 'Remove spaces after std_logic_vector'

    def _get_tokens_of_interest(self, oFile):
        return [oFile.get_all_tokens()]

    def _analyze(self, lToi):
        oToi = lToi[0]
        iLine, lTokens = utils.get_toi_parameters(oToi)
        for iToken, oToken in enumerate(lTokens[:len(lTokens) - 2]):

            iLine = utils.increment_line_number(iLine, oToken)

            if oToken.get_value().lower() == 'std_logic_vector':
                if utils.are_next_consecutive_token_types([parser.whitespace, parser.open_parenthesis], iToken + 1, lTokens):
                    lExtractedTokens = oToi.extract_tokens(iToken, iToken + 1)
                    oViolation = violation.New(iLine, lExtractedTokens, self.solution)
                    self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        lTokens.pop()
        oViolation.set_tokens(lTokens)




from vsg import parser


class entity_keyword(parser.keyword):
    '''
    unique_id = entity_aspect : entity_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class entity_name(parser.name):
    '''
    unique_id = entity_aspect : entity_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class open_parenthesis(parser.open_parenthesis):
    '''
    unique_id = process_statement : open_parenthesis
    '''

    def __init__(self, sString='('):
        parser.open_parenthesis.__init__(self)


class architecture_identifier(parser.identifier):
    '''
    unique_id = entity_aspect : architecture_identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class close_parenthesis(parser.close_parenthesis):
    '''
    unique_id = process_statement : close_parenthesis
    '''

    def __init__(self, sString=')'):
        parser.close_parenthesis.__init__(self)


class configuration_keyword(parser.keyword):
    '''
    unique_id = entity_aspect : configuration_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class configuration_name(parser.name):
    '''
    unique_id = entity_aspect : configuration_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class open_keyword(parser.keyword):
    '''
    unique_id = entity_aspect : open_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg import parser
from vsg import violation

from vsg.rule_group import structure
from vsg.vhdlFile import utils


class remove_lines_starting_with_token_between_token_pairs(structure.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token type pairs
       The tokens to check for a single space between
    '''

    def __init__(self, name, identifier, oRemoveToken, lTokens):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 2
        self.oRemoveToken = oRemoveToken
        self.lTokens = lTokens

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokens:
            lToi_a = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = utils.combine_two_token_class_lists(lToi, lToi_a)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, parser.carriage_return):
                    iLine += 1
                    if utils.are_next_consecutive_token_types([parser.whitespace, self.oRemoveToken], iToken + 1, lTokens):
                        oSubToi = oToi.extract_tokens(iToken, iToken + 2)
                        oViolation = violation.New(iLine, oSubToi, self.solution)
                        self.add_violation(oViolation)
                    if utils.are_next_consecutive_token_types([self.oRemoveToken], iToken + 1, lTokens):
                        oSubToi = oToi.extract_tokens(iToken, iToken + 1)
                        oViolation = violation.New(iLine, oSubToi, self.solution)
                        self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        oViolation.set_tokens([])




from vsg.rules import blank_line_above_line_starting_with_token

from vsg.token import port_clause as token

lTokens = []
lTokens.append(token.port_keyword)


class rule_001(blank_line_above_line_starting_with_token):
    '''
    This rule checks for a blank line above the **port** keyword.

    **Violation**

    .. code-block:: vhdl

       entity FIFO is

         port (

    **Fix**

    .. code-block:: vhdl

       entity FIFO is
         port (
    '''
    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'port', '001', lTokens)
        self.style = 'no_blank_line'




from vsg.rules import token_case as Rule

from vsg import token

lTokens = []
lTokens.append(token.iteration_scheme.for_keyword)


class rule_501(Rule):
    '''
    This rule checks the **while** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       FOR x in (31 downto 0) loop

    **Fix**

    .. code-block:: vhdl

       for x in (31 downto 0) loop
    '''

    def __init__(self):
        Rule.__init__(self, 'iteration_scheme', '501', lTokens)
        self.groups.append('case::keyword')




from vsg.token import incomplete_type_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import identifier


def detect(iToken, lObjects):
    '''
    incomplete_type_declaration ::=
        type identifier ;
    '''
    if utils.are_next_consecutive_tokens(['type', None, ';'], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('type', token.type_keyword, iToken, lObjects)

    iCurrent = identifier.classify(iCurrent, lObjects, token.identifier)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import token

from vsg.rules import token_case

lTokens = []
lTokens.append(token.case_generate_statement.end_generate_keyword)
lTokens.append(token.for_generate_statement.end_generate_keyword)
lTokens.append(token.if_generate_statement.end_generate_keyword)


class rule_010(token_case):
    '''
    This rule checks the **generate** keyword has the proper case in the **end generate** line.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end GENERATE ram_array;

    **Fix**

    .. code-block:: vhdl

       end generate ram_array;
    '''

    def __init__(self):
        token_case.__init__(self, 'generate', '010', lTokens)
        self.groups.append('case::keyword')




from vsg import deprecated_rule


class rule_026(deprecated_rule.Rule):
    '''
    The function of this rule has been moved to whitespace_200.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'context', '026')
        self.message.append('Rule ' + self.unique_id + ' has been moved to whitespace_200.')




from vsg.rules import remove_excessive_blank_lines_below_line_ending_with_token

from vsg.token import port_clause as token

lTokens = []
lTokens.append(token.open_parenthesis)


class rule_022(remove_excessive_blank_lines_below_line_ending_with_token):
    '''
    This rule checks for blank lines after the **port** keyword.

    **Violation**

    .. code-block:: vhdl

       port (


         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW: out   std_logic;
         DATA     : inout std_logic
       );

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic
       );
    '''
    def __init__(self):
        remove_excessive_blank_lines_below_line_ending_with_token.__init__(self, 'port', '022', lTokens, iAllow=0)
        self.solution = 'Remove blank lines below *port* keyword.'




from vsg.rules import token_prefix

from vsg import token

lTokens = []
lTokens.append(token.variable_declaration.identifier)


class rule_012(token_prefix):
    '''
    This rule checks for valid prefixes on variable identifiers.
    The default variable prefix is *v\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       variable my_var : natural;

    **Fix**

    .. code-block:: vhdl

       variable v_my_var : natural;
    '''

    def __init__(self):
        token_prefix.__init__(self, 'variable', '012', lTokens)
        self.prefixes = ['v_']
        self.solution = 'Variable identifiers'




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import component_declaration as token


class rule_013(Rule):
    '''
    This rule checks for a single space after the **component** keyword in the **end component** line.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end component    fifo;

    **Fix**

    .. code-block:: vhdl

       end component fifo;
    '''
    def __init__(self):
        Rule.__init__(self, 'component', '013')
        self.left_token = token.end_component_keyword
        self.right_token = token.component_simple_name




from vsg.rules import token_indent as Rule

from vsg import token

lTokens = []
lTokens.append(token.alias_declaration.alias_keyword)


class rule_300(Rule):
    '''
    This rule checks the indent of the **alias** keyword.

    **Violation**

    .. code-block:: vhdl

       signal sig1 : integer;

         alias is name;

    **Fix**

    .. code-block:: vhdl

       signal sig1 : integer;

       alias is name;
    '''

    def __init__(self):
        Rule.__init__(self, 'alias_declaration', '300', lTokens)




from vsg.vhdlFile.classify import block_statement
from vsg.vhdlFile.classify import component_instantiation_statement
from vsg.vhdlFile.classify import concurrent_assertion_statement
from vsg.vhdlFile.classify import concurrent_procedure_call_statement
from vsg.vhdlFile.classify import concurrent_signal_assignment_statement
from vsg.vhdlFile.classify import generate_statement
from vsg.vhdlFile.classify import process_statement


def detect(iToken, lObjects):
    '''
    concurrent_statement ::=
        block_statement
      | process_statement
      | concurrent_procedure_call_statement
      | concurrent_assertion_statement
      | concurrent_signal_assignment_statement
      | component_instantiation_statement
      | generate_statement
      | PSL_PSL_Directive
    '''

    iCurrent = process_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = block_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = generate_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = concurrent_assertion_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = concurrent_signal_assignment_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = concurrent_procedure_call_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = component_instantiation_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken




from vsg import parser


class identifier(parser.identifier):
    '''
    unique_id = secondary_unit_declaration : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class equal_sign(parser.item):
    '''
    unique_id = secondary_unit_declaration : equal_sign
    '''

    def __init__(self):
        parser.item.__init__(self, '=')


class physical_literal(parser.item):
    '''
    unique_id = secondary_unit_declaration : physical_literal
    '''

    def __init__(self, sString):
        parser.item.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = secondary_unit_declaration : semicolon
    '''

    def __init__(self):
        parser.semicolon.__init__(self)




from vsg import parser


class constant_keyword(parser.keyword):
    '''
    unique_id = interface_constant_declaration : constant_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class identifier(parser.identifier):
    '''
    unique_id = interface_constant_declaration : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class subtype_indication(parser.subtype_indication):
    '''
    unique_id = interface_constant_declaration : subtype_indication
    '''
    def __init__(self, sString):
        parser.subtype_indication.__init__(self, sString)


class colon(parser.colon):
    '''
    unique_id = interface_constant_declaration : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)


class assignment(parser.assignment):
    '''
    unique_id = interface_constant_declaration : assignment
    '''

    def __init__(self, sString):
        parser.assignment.__init__(self, sString)




from vsg import parser
from vsg import token
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import structure

lSplitTokens = []
lSplitTokens.append(token.conditional_waveforms.else_keyword)

lTokenPairs = []
lTokenPairs.append([token.concurrent_conditional_signal_assignment.assignment, token.concurrent_conditional_signal_assignment.semicolon])


class rule_007(structure.Rule):
    '''
    This rule checks for code after the **else** keyword.

    .. NOTE:: There is a configuration option **allow_single_line** which allows single line concurrent statements.

    allow_single_line set to False (Default)
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Violation**

    .. code-block:: vhdl

       wr_en <= '0' when overflow = '0' else '1';
       wr_en <= '0' when overflow = '0' else '1' when underflow = '1' else sig_a;

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0' when overflow = '0' else
                '1';
       wr_en <= '0' when overflow = '0' else
                '1' when underflow = '1' else
                sig_a;

    allow_single_line set to True
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Violation**

    .. code-block:: vhdl

       wr_en <= '0' when overflow = '0' else '1';
       wr_en <= '0' when overflow = '0' else '1' when underflow = '1' else sig_a;

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0' when overflow = '0' else '1';
       wr_en <= '0' when overflow = '0' else
                '1' when underflow = '1' else
                sig_a;
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'concurrent', '007')
        self.solution = 'move code after else to next line.'
        self.subphase = 2
        self.lSplitTokens = lSplitTokens
        self.lTokenPairs = lTokenPairs
        self.allow_single_line = False
        self.configuration.append('allow_single_line')

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = utils.combine_two_token_class_lists(lToi, aToi)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()

            if utils.find_carriage_return(lTokens) is None and self.allow_single_line:
                for oSplitToken in self.lSplitTokens:
                    if utils.count_token_types_in_list_of_tokens(oSplitToken, lTokens) > 1:
                        break
                else:
                    continue

            iLine = oToi.get_line_number()
            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)
                for oSplitToken in self.lSplitTokens:
                    if isinstance(oToken, oSplitToken):
                        if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], iToken + 1, lTokens):
                            continue
                        if utils.are_next_consecutive_token_types([parser.comment], iToken + 1, lTokens):
                            continue
                        if utils.are_next_consecutive_token_types([parser.carriage_return], iToken + 1, lTokens):
                            continue
                        oViolation = violation.New(iLine, oToi.extract_tokens(iToken, iToken), self.solution)
                        self.add_violation(oViolation)
                        break

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        lTokens.append(parser.carriage_return())
        oViolation.set_tokens(lTokens)




from vsg.token import process_sensitivity_list as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import sensitivity_list


def classify(iToken, lObjects):
    '''
    process_sensitivity_list ::=
        all | sensitivity_list
    '''

    if utils.is_next_token('all', iToken, lObjects):
        return utils.assign_next_token_required('all', token.all_keyword, iToken, lObjects)
    else:
        return sensitivity_list.classify(iToken, lObjects)




from vsg.rules import token_case_in_range_bounded_by_tokens

from vsg import token

lTokens = []
lTokens.append(token.subprogram_body.is_keyword)

oStartToken = token.function_specification.function_keyword
oEndToken = token.subprogram_body.semicolon


class rule_502(token_case_in_range_bounded_by_tokens):
    '''
    This rule checks the **is** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       function overflow (a: integer) return integer IS

    **Fix**

    .. code-block:: vhdl

       function overflow (a: integer) return integer is
    '''

    def __init__(self):
        token_case_in_range_bounded_by_tokens.__init__(self, 'function', '502', lTokens, oStartToken, oEndToken)
        self.groups.append('case::keyword')




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment

from vsg.token import generic_map_aspect as token

lTokens = []
lTokens.append(token.open_parenthesis)


class rule_005(insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment):
    '''
    This rule checks if the **generic map** keywords and a generic assignment are on the same line.

    **Violation**

    .. code-block:: vhdl

         generic map (DEPTH => 512,
           WIDTH => 32
         )

    **Fix**

    .. code-block:: vhdl

         generic map (
           DEPTH => 512,
           WIDTH => 32
         )
    '''

    def __init__(self):
        insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment.__init__(self, 'generic_map', '005', lTokens)
        self.solution = 'Move generic assignment to it\'s own line.'




from vsg.rules import token_suffix

from vsg import token

lTokens = []
lTokens.append(token.loop_statement.loop_label)


class rule_601(token_suffix):
    '''
    This rule checks for valid suffixes on loop labels.
    The default prefix is *\_loop*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       label : for index in 4 to 23 loop

    **Fix**

    .. code-block:: vhdl

       label_loop : for index in 4 to 23 loop
    '''

    def __init__(self):
        token_suffix.__init__(self, 'loop_statement', '601', lTokens)
        self.suffixes = ['_loop']
        self.solution = 'Loop labels'




from vsg import parser


class identifier(parser.identifier):
    '''
    unique_id = parameter_specification : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class in_keyword(parser.keyword):
    '''
    unique_id = parameter_specification : in_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg.rules import previous_line

from vsg.token import architecture_body as token


class rule_003(previous_line):
    '''
    This rule checks for a blank lines or comments above the **architecture** declaration.

    |configuring_previous_line_rules_link|

    **Violation**

    .. code-block:: vhdl

       library ieee;
       architecture rtl of fifo is

    **Fix**

    .. code-block:: vhdl

       library ieee;

       architecture rtl of fifo is
    '''

    def __init__(self):
        previous_line.__init__(self, 'architecture', '003', [token.architecture_keyword])
        self.style = 'require_blank_line'




import json
import hashlib


def write(commandLineArguments, dJson):

    lReport = build_report(dJson)
    write_report_to_file(lReport, commandLineArguments)


def build_report(dJson):
    lReport = []

    iViolation = 0
    for dFile in dJson['files']:
        for dViolation in dFile['violations']:
            iViolation += 1
            dEntry = create_entry(dFile, dViolation, iViolation)
            lReport.append(dEntry)
    return lReport


def write_report_to_file(lReport, commandLineArguments):
    with open(commandLineArguments.quality_report, 'w') as oFile:
        oFile.write(json.dumps(lReport, indent=2))


def create_entry(dFile, dViolation, iViolation):
    dReturn = {}
    dReturn['description'] = build_description(dViolation)
    dReturn['fingerprint'] = build_fingerprint(dFile, dViolation, iViolation)
    dReturn['severity'] = remap_severity(dViolation['severity'])
    dReturn['location'] = build_location(dFile, dViolation)
    return dReturn


def build_description(dViolation):
    return dViolation['rule'] + ' :: ' + dViolation['solution']


def build_fingerprint(dFile, dViolation, iViolation):
    sHashString = build_description(dViolation)
    sHashString += remap_severity(dViolation['severity'])
    sHashString += dFile['file_path']
    sHashString += str(dViolation['linenumber'])
    sHashString += str(iViolation)
    sHash = hashlib.md5(sHashString.encode('utf-8')).hexdigest()
    return sHash


def build_location(dFile, dViolation):
    dReturn = {}
    dReturn['path'] = dFile['file_path']
    dReturn['lines'] = {}
    dReturn['lines']['begin'] = dViolation['linenumber']
    return dReturn


def remap_severity(sSeverity):
    if sSeverity == 'Error':
        return 'critical'
    else:
        return 'minor'




from vsg.rules import move_token_next_to_another_token

from vsg.token import generic_map_aspect as token


class rule_003(move_token_next_to_another_token):
    '''
    This rule checks the ( is on the same line as the **generic map** keywords.

    **Violation**

    .. code-block:: vhdl

       generic map
       (
         WIDTH => 32,
         DEPTH => 512
       )

    **Fix**

    Use explicit port mapping.

    .. code-block:: vhdl

       generic map (
         WIDTH => 32,
         DEPTH => 512
       )
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'generic_map', '003', token.map_keyword, token.open_parenthesis)
        self.solution = 'Move the ( to the same line as the *generic map* keywords.'




from vsg import token

from vsg.rules import token_case_formal_part_of_association_element_in_map_between_tokens

oStart = token.component_instantiation_statement.instantiation_label
oEnd = token.component_instantiation_statement.semicolon

sMapType = 'generic'


class rule_002(token_case_formal_part_of_association_element_in_map_between_tokens):
    '''
    This rule checks generic names have proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

         generic map (
           DEPTH => 512,
           WIDTH => 32
         )

    **Fix**

    .. code-block:: vhdl

         generic map (
           depth => 512,
           width => 32
         )
    '''
    def __init__(self):
        token_case_formal_part_of_association_element_in_map_between_tokens.__init__(self, 'generic_map', '002', sMapType, oStart, oEnd)
        self.configuration.append('prefix_exceptions')
        self.configuration.append('suffix_exceptions')
        self.configuration.append('case_exceptions')
        self.groups.append('case::name')




from vsg.token import simple_release_assignment as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import force_mode


def detect(iToken, lObjects):
    '''
    simple_release_assignment ::=
        target <= release [ force_mode ] ;
    '''

    if utils.find_in_range('release', iToken, ';', lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_tokens_until('<=', token.target, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('<=', token.assignment, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('release', token.release_keyword, iCurrent, lObjects)

    iCurrent = force_mode.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.token import instantiation_list as token

from vsg.vhdlFile import utils


def classify(iToken, lObjects):
    '''
    instantiation_list ::=
        instantiation_label { , instantiation_label }
      | **others**
      | **all**
    '''

    if utils.is_next_token('others', iToken, lObjects):
        return utils.assign_next_token_required('others', token.others_keyword, iToken, lObjects)

    if utils.is_next_token('all', iToken, lObjects):
        return utils.assign_next_token_required('all', token.all_keyword, iToken, lObjects)

    iCurrent = utils.assign_next_token(token.instantiation_label, iToken, lObjects)

    while utils.is_next_token(',', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required(',', token.comma, iToken, lObjects)
        iCurrent = utils.assign_next_token(token.instantiation_label, iToken, lObjects)

    return iCurrent




from vsg.rules import split_line_at_token_when_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.report_statement.severity_keyword)

oStart = token.report_statement.report_keyword
oEnd = token.report_statement.semicolon


class rule_002(split_line_at_token_when_between_tokens):
    '''
    This rule checks the **severity** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

        report "FIFO width is limited to 16 bits." severity FAILURE;

    **Fix**

    .. code-block:: vhdl

        report "FIFO width is limited to 16 bits."
          severity FAILURE;
    '''

    def __init__(self):
        split_line_at_token_when_between_tokens.__init__(self, 'report_statement', '002', lTokens, oStart, oEnd)
        self.solution = "Place **severity** keyword on its own line."




from vsg.rules import split_line_at_token

from vsg import token

lTokens = []
lTokens.append(token.library_clause.keyword)


class rule_010(split_line_at_token):
    '''
    This rule checks the **library** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       context c1 is library ieee; use ieee.std_logic_1164.all; end context c1;

    **Fix**

    .. code-block:: vhdl

       context c1 is
         library ieee; use ieee.std_logic_1164.all; end context c1;
    '''

    def __init__(self):
        split_line_at_token.__init__(self, 'library', '010', lTokens)
        self.solution = 'Move *library* to it\'s own line.'




from vsg.rules import token_indent_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.identifier_list.identifier)

oStart = token.record_type_definition.record_keyword
oEnd = token.record_type_definition.end_keyword

class rule_012(token_indent_between_tokens):
    '''
    This rule checks the indent of record elements in record type declarations.

    **Violation**

    .. code-block:: vhdl

       type interface is record
         data : std_logic_vector(31 downto 0);
       chip_select : std_logic;
           wr_en : std_logic;
       end record;

    **Fix**

    .. code-block:: vhdl

       type interface is record
         data : std_logic_vector(31 downto 0);
         chip_select : std_logic;
         wr_en : std_logic;
       end record;
    '''

    def __init__(self):
        token_indent_between_tokens.__init__(self, 'type', '012', lTokens, oStart, oEnd)




from vsg import violation

from vsg.rule_group import structure
from vsg.vhdlFile import utils


class remove_tokens(structure.Rule):
    '''
    Removes a token and duplicate whitespace.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : token object type list
       tokens to remove

    '''

    def __init__(self, name, identifier, lTokens):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 1
        self.lTokens = lTokens

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_n_tokens_before_and_after_tokens(1, self.lTokens)

    def _analyze(self, lToi):
        for oToi in lToi:
           self.add_violation(violation.New(oToi.get_line_number(), oToi, self.solution))

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        lTokens.pop(1)
        lTokens = utils.remove_consecutive_whitespace_tokens(lTokens)
        oViolation.set_tokens(lTokens)




from vsg.token import interface_incomplete_type_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import identifier


def detect(iToken, lObjects):
    '''
    interface_incomplete_type_declaration ::=
        type identifier
    '''
    if utils.is_next_token('type', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_if('type', token.type_keyword, iToken, lObjects)
    iCurrent = identifier.classify(iCurrent, lObjects)

    return iCurrent




from vsg.token import index_constraint as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import discrete_range


def detect(iToken, lObjects):
    '''
    index_constraint ::=
        ( discrete_range { , discrete_range } )
    '''
    if utils.is_next_token('(', iToken, lObjects):
        iCurrent = utils.find_next_token(iToken, lObjects) + 1
        if discrete_range.detect(iCurrent, lObjects):
            return True
    return False


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iToken, lObjects)

    while not utils.is_next_token(')', iCurrent, lObjects):
        iCurrent = discrete_range.classify_until([','], iCurrent, lObjects)
        iCurrent = utils.assign_next_token_if(',', token.comma, iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)
    return iCurrent




from vsg.rules import remove_comments_from_end_of_lines_bounded_by_tokens

from vsg import token

oStart = token.component_instantiation_statement.instantiation_label

oEnd = token.component_instantiation_statement.semicolon


class rule_023(remove_comments_from_end_of_lines_bounded_by_tokens):
    '''
    This rule checks for comments at the end of the port and generic assignments in instantiations.
    These comments represent additional maintainence.
    They will be out of sync with the entity at some point.
    Refer to the entity for port types, port directions and purpose.

    **Violation**

    .. code-block:: vhdl

       WR_EN => w_wr_en;   -- out : std_logic
       RD_EN => w_rd_en;   -- Reads data when asserted

    **Fix**

    .. code-block:: vhdl

       WR_EN => w_wr_en;
       RD_EN => w_rd_en;
    '''

    def __init__(self):
        remove_comments_from_end_of_lines_bounded_by_tokens.__init__(self, 'instantiation', '023', oStart, oEnd)
        self.solution = 'Remove comment.'




from vsg import parser
from vsg import token

from vsg.vhdlFile.extract import tokens


def get_interface_elements_between_tokens(oStart, oEnd, lAllTokens, oTokenMap):
    lReturn = []
    lStartIndexes, lEndIndexes = oTokenMap.get_token_pair_indexes(oStart, oEnd)

    for iStart, iEnd in zip(lStartIndexes, lEndIndexes):
        iLine = oTokenMap.get_line_number_of_index(iStart)
        lToi = lAllTokens[iStart + 1:iEnd + 1]

        bStore = False
        iLineNumber = None
        lTemp = []
        for iIndex in range(0, len(lToi)):
            oToken = lToi[iIndex]
            if not isinstance(oToken, parser.whitespace) and not isinstance(oToken, parser.carriage_return) and not isinstance(oToken, parser.comment) and not bStore:
                bStore = True
                iStartIndex = iIndex + iStart + 1
                iLineNumber = iLine

            if bStore:
               lTemp.append(lToi[iIndex])

            if isinstance(oToken, token.interface_list.semicolon):
                lTemp.pop()
                lReturn.append(tokens.New(iStartIndex, iLineNumber, lTemp))
                lTemp = []
                bStore = False

            if isinstance(lToi[iIndex], parser.carriage_return):
                iLine +=1

        if len(lTemp) > 0:
            lTemp.pop()
            for i in range(1, 5):
                if isinstance(lTemp[-1], parser.whitespace):
                    lTemp.pop()
                    continue
                if isinstance(lTemp[-1], parser.carriage_return):
                    lTemp.pop()
                    continue
                if isinstance(lTemp[-1], parser.comment):
                    lTemp.pop()
                    continue
            lReturn.append(tokens.New(iStartIndex, iLineNumber, lTemp))

    return lReturn




from vsg import token

from vsg.rules.whitespace_after_token import Rule

lTokens = []
lTokens.append(token.signal_declaration.colon)


class rule_005(Rule):
    '''
    This rule checks for a single space after the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       signal wr_en :    std_logic;
       signal rd_en :std_logic;

    **Fix**

    .. code-block:: vhdl

       signal wr_en : std_logic;
       signal rd_en : std_logic;
    '''
    def __init__(self):
        Rule.__init__(self, 'signal', '005', lTokens)




from vsg import parser

from vsg.vhdlFile.extract import tokens

import bisect


def get_line_preceeding_line(iLine, lAllTokens, iNumLines, oTokenMap, bSkipComments=False):

    lCarriageReturns = oTokenMap.get_token_indexes(parser.carriage_return)

    if not bSkipComments:
        iAdjust = -2
        iStartIndex = iLine - iNumLines + iAdjust
        if iStartIndex < 0:
            iStart = 0
        else:
            iStart = lCarriageReturns[iStartIndex] + 1
        iEnd = lCarriageReturns[iLine + iAdjust]
        lTemp = lAllTokens[iStart:iEnd]
    #    print(f'{iLine} | {iStart} | {iEnd} | {lTemp}')
        return tokens.New(iStart, iLine, lTemp)
    else:

        iStartIndex = _get_start_index(lCarriageReturns, iLine, oTokenMap)

        if iStartIndex == 0:
            iTokenStartIndex = 0
            iTokenEndIndex = lCarriageReturns[iTokenStartIndex]
        else:
            iTokenStartIndex = lCarriageReturns[iStartIndex - 1] + 1
            iTokenEndIndex = lCarriageReturns[iStartIndex]

        iLine = iStartIndex + 2
        lTokens = lAllTokens[iTokenStartIndex:iTokenEndIndex]
        return tokens.New(iTokenStartIndex, iLine, lTokens)


def _build_index_list(oTokenMap):
        lTemp = oTokenMap.get_token_indexes(parser.comment).copy()
        lTemp.extend(oTokenMap.get_token_indexes(parser.whitespace).copy())
        lTemp.extend(oTokenMap.get_token_indexes(parser.carriage_return).copy())
        lTemp.sort()
        return lTemp


def _get_start_index(lCarriageReturns, iLine, oTokenMap):
        lTemp = _build_index_list(oTokenMap)
        iCurrent = lCarriageReturns[iLine - 2]
#        print(f'lCarraigeReturns = {lCarriageReturns}')
#        print(f'Index = {iCurrent}')
#        print(f'lTemp            = {lTemp}')
        iTempIndex = lTemp.index(iCurrent) - 1
#        print(f'iTempIndex = {iTempIndex}')
        for i in range(iTempIndex, -1, -1):
#            print(i)
#            print(f'lTemp[i] | {iCurrent - 1}')
            if lTemp[i] != iCurrent - 1:
                break
            iCurrent = lTemp[i]
#        print(f'iCurrent = {iCurrent}')
#        iStartIndex = lCarriageReturns.index(iCurrent)
        iStartIndex = bisect.bisect_left(lCarriageReturns, iCurrent)
        return iStartIndex




from vsg.rule_group import case
from vsg import violation

lKeywords = []
lKeywords.append('std_logic')
lKeywords.append('std_logic_vector')
lKeywords.append('integer')
lKeywords.append('signed')
lKeywords.append('unsigned')
lKeywords.append('natural')
lKeywords.append('std_ulogic')


class token_case_n_token_after_tokens_between_tokens(case.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    trigger : parser object type
       object type to apply the case check against
    '''

    def __init__(self, name, identifier, iToken, lTokens, oStart, oEnd, bLimitToVhdlKeywords=False):
        case.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 6
        self.case = 'lower'
        self.configuration.append('case')
        self.iToken = iToken
        self.lTokens = lTokens
        self.oStart = oStart
        self.oEnd = oEnd
        self.disabled = False
        self.bLimitToVhdlKeywords = bLimitToVhdlKeywords

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_n_token_after_tokens_between_tokens(self.iToken, self.lTokens, self.oStart, self.oEnd)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            sObjectValue = lTokens[0].get_value()
            if self.bLimitToVhdlKeywords:
                if sObjectValue.lower() not in lKeywords:
                    continue
            if self.case == 'lower':
                if not sObjectValue.islower():
                    sSolution = 'Change "' + sObjectValue + '" to "' + sObjectValue.lower() + '"'
                    self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))
            if self.case == 'upper':
                if not sObjectValue.isupper():
                    sSolution = 'Change "' + sObjectValue + '" to "' + sObjectValue.upper() + '"'
                    self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if self.case == 'lower':
            lTokens[0].set_value(lTokens[0].get_value().lower())
        if self.case == 'upper':
            lTokens[0].set_value(lTokens[0].get_value().upper())
        oViolation.set_tokens(lTokens)




from vsg.vhdlFile.classify import interface_list


def classify(iToken, lObjects):
    '''
    formal_parameter_list ::=
        *parameter*_interface_list
    '''

    return interface_list.classify(iToken, lObjects)




from vsg.token import component_configuration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import component_specification
from vsg.vhdlFile.classify import binding_indication
from vsg.vhdlFile.classify import block_configuration


def detect(iToken, lObjects):

    if utils.is_next_token('for', iToken, lObjects):
        iCurrent = utils.find_next_token(iToken, lObjects) + 1
        if component_specification.detect(iCurrent, lObjects):
            return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):
    '''
    component_configuration ::=
        for component_specification
            [ binding_indication ; ]
            { verification_unit_binding_indication ; }
            [ block_configuration ]
        end for ;
    '''

    iCurrent = utils.assign_next_token_required('for', token.for_keyword, iToken, lObjects)

    iCurrent = component_specification.classify(iCurrent, lObjects)

    iPrevious = iCurrent
    iCurrent = binding_indication.detect(iCurrent, lObjects)
    if not iPrevious == iCurrent:
        iCurrent = utils.assign_next_token_required(';', token.binding_indication_semicolon, iCurrent, lObjects)

    iCurrent = block_configuration.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('for', token.end_for_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_case

from vsg.token import architecture_body as token


class rule_009(token_case):
    '''
    This rule checks the **end** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       END architecture;

       End architecture;

    **Fix**

    .. code-block:: vhdl

       end architecture;

       end architecture;
    '''

    def __init__(self):
        token_case.__init__(self, 'architecture', '009', [token.end_keyword])
        self.groups.append('case::keyword')




from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_token_and_n_tokens_after_it_when_between_tokens(lTokens, iTokens, oStart, oEnd, lAllTokens, oTokenMap):
    lReturn = []

    lIndexes = utils.filter_tokens_between_tokens(lTokens, oStart, oEnd, oTokenMap)

    for iIndex in lIndexes:
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        iEnd = iIndex + iTokens + 1
        lReturn.append(tokens.New(iIndex, iLine, lAllTokens[iIndex:iEnd]))

    return lReturn




from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_sequence_of_tokens_matching(lTokens, lAllTokens, oTokenMap, bIgnoreIfLineStart=False):

    lReturn = []

    lIndexes = oTokenMap.get_token_indexes(lTokens[0])

    for iIndex in lIndexes:
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        if bIgnoreIfLineStart and utils.is_token_at_start_of_line(iIndex, oTokenMap):
            continue
        for iToken, oToken in enumerate(lTokens):
            if not isinstance(lAllTokens[iToken + iIndex], oToken):
                break
        else:
            lReturn.append(tokens.New(iIndex, iLine, lAllTokens[iIndex:iIndex + len(lTokens)]))

    return lReturn




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.component_instantiation_statement.instantiation_label, token.component_instantiation_statement.label_colon])


class rule_003(Rule):
    '''
    This rule checks for a single space before the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       U_FIFO: FIFO

    **Fix**

    .. code-block:: vhdl

       U_FIFO : FIFO
    '''
    def __init__(self):
        Rule.__init__(self, 'instantiation', '003', lTokens)




from vsg.rules import blank_line_below_line_ending_with_token

from vsg.token import architecture_body as token


class rule_017(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **begin** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       begin
         wr_en <= '0';

    **Fix**

    .. code-block:: vhdl

       begin

         wr_en <= '0';
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'architecture', '017', [token.begin_keyword])




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.assertion.keyword)
lTokens.append(token.assertion.report_keyword)
lTokens.append(token.assertion.severity_keyword)
lTokens.append(token.concurrent_assertion_statement.label_name)
lTokens.append(token.assertion_statement.label)


class rule_001(token_indent):
    '''
    This rule checks indent of multiline assert statements.

    **Violation**

    .. code-block:: vhdl

       assert WIDTH > 16
            report "FIFO width is limited to 16 bits."
        severity FAILURE;

    **Fix**

    .. code-block:: vhdl

       assert WIDTH > 16
         report "FIFO width is limited to 16 bits."
         severity FAILURE;
    '''

    def __init__(self):
        token_indent.__init__(self, 'assert', '001', lTokens)




from vsg import deprecated_rule


class rule_004(deprecated_rule.Rule):
    '''
    This rule has been split into the following rules:

    * :ref:`context_ref_500`
    * :ref:`context_ref_501`
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'context_ref', '004')
        self.message.append('Rule ' + self.unique_id + ' has been split into the following rules:')
        self.message.append('  context_ref_500')
        self.message.append('  context_ref_501')




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class insert_token_right_of_possible_tokens_if_it_does_not_exist_before_token(structure.Rule):
    '''
    Checks for the existence of a token and will insert it if it does not exist.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    insert_token : token object
       token to insert if it does not exist.

    anchor_token : token object type
       token to check if insert_token exists to the right of

    end_token : token object type
       token that bounds the search for the insert_token
    '''

    def __init__(self, name, identifier, insert_token, anchor_tokens, end_token):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.oInsertToken = insert_token
        self.lAnchorTokens = anchor_tokens
        self.oEndToken = end_token
        self.action = 'add'
        self.configuration.append('action')

    def _get_tokens_of_interest(self, oFile):
        if self.action == 'add':
            return oFile.get_tokens_bounded_by(self.lAnchorTokens[0], self.oEndToken)
        else:
            return oFile.get_token_and_n_tokens_before_it([self.oInsertToken], 1)

    def _analyze(self, lToi):
        if self.action == 'remove':
            for oToi in lToi:
                sSolution = self.action.capitalize() + ' ' + self.solution
                self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))
            return

        for oToi in lToi:

            iLine, lTokens = utils.get_toi_parameters(oToi)

            if utils.does_token_type_exist_in_list_of_tokens(type(self.oInsertToken), lTokens):
                continue

            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)
                for oSearch in self.lAnchorTokens:
                    if isinstance(oToken, oSearch):
                        iIndex = iToken
                        iLineNumber = iLine

            sSolution = self.action.capitalize() + ' ' + self.solution
            dAction = {}
            dAction['whitespace'] = False
            dAction['carriage_return'] = False
            if isinstance(lTokens[iIndex + 1], parser.whitespace):
                dAction['whitespace'] = True
            if isinstance(lTokens[iIndex + 1], parser.carriage_return):
                dAction['carriage_return'] = True

            oViolation = violation.New(iLineNumber, oToi.extract_tokens(iIndex, iIndex), sSolution)
            oViolation.set_action(dAction)
            self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if self.action == 'remove':
            rules_utils.remove_optional_item(oViolation, self.oInsertToken)
        else:
            dAction = oViolation.get_action()
            if isinstance(lTokens[0], parser.close_parenthesis) and dAction['carriage_return']:
                rules_utils.insert_token(lTokens, 1, self.oInsertToken)
            elif isinstance(lTokens[0], parser.close_parenthesis) and not dAction['whitespace']:
                rules_utils.insert_whitespace(lTokens, 1)
                rules_utils.insert_token(lTokens, 1, self.oInsertToken)
            else:
                rules_utils.insert_token(lTokens, 1, self.oInsertToken)
                rules_utils.insert_whitespace(lTokens, 1)
            oViolation.set_tokens(lTokens)




from vsg.rules import previous_line

from vsg.token import function_specification as token

lTokens = []
lTokens.append(token.function_keyword)
lTokens.append(token.pure_keyword)
lTokens.append(token.impure_keyword)


class rule_006(previous_line):
    '''
    This rule checks for blank lines or comments above the **function** keyword.

    |configuring_previous_line_rules_link|

    **Violation**

    .. code-block:: vhdl

       architecture RTL of FIFO is
         function overflow (a: integer) return integer is


    **Fix**

    .. code-block:: vhdl

       architecture RTL of FIFO is

         function overflow (a: integer) return integer is
    '''

    def __init__(self):
        previous_line.__init__(self, 'function', '006', lTokens)




from vsg.token import type_mark as token

from vsg.vhdlFile import utils


def classify(iToken, lObjects):
    '''
    type_mark ::=
        *type*_name
      | *subtype*_name
    '''

    return utils.assign_next_token(token.name, iToken, lObjects)





from vsg import parser
from vsg import violation

from vsg.rule_group import structure
from vsg.rules import utils as rules_utils


class split_line_at_token_when_between_tokens_unless_token_is_found(structure.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of parser object types
       object type to split a line at

    oStart : token type
       The start of the range

    oEnd : token type
       The end of the range
    '''

    def __init__(self, name, identifier, lTokens, oStart, oEnd, oStop):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 1
        self.lTokens = lTokens
        self.oStart = oStart
        self.oEnd = oEnd
        self.oStop = oStop

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_token_and_n_tokens_before_it_in_between_tokens_unless_token_is_found(self.lTokens, 2, self.oStart, self.oEnd, self.oStop)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if isinstance(lTokens[0], parser.carriage_return) or isinstance(lTokens[1], parser.carriage_return):
                continue
            sSolution = self.solution
            oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
            self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if isinstance(lTokens[1], parser.whitespace):
            rules_utils.insert_carriage_return(lTokens, -2)
        else:
            rules_utils.insert_carriage_return(lTokens, -1)
        oViolation.set_tokens(lTokens)




from vsg.token import wait_statement as token

from vsg.vhdlFile.classify import sensitivity_clause
from vsg.vhdlFile.classify import condition_clause
from vsg.vhdlFile.classify import timeout_clause

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    wait_statement ::=
        [ label : ] wait [ sensitivity_clause ] [ condition_clause ] [ timeout_clause ] ;
    '''
    if utils.keyword_found('wait', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.label, token.label_colon)
    iCurrent = utils.assign_next_token_required('wait', token.wait_keyword, iCurrent, lObjects)

    if sensitivity_clause.detect(iCurrent, lObjects):
        iCurrent = sensitivity_clause.classify_until([';', 'for', 'until'], iCurrent, lObjects)

    if condition_clause.detect(iCurrent, lObjects):
        iCurrent = condition_clause.classify_until([';', 'for'], iCurrent, lObjects)

    if timeout_clause.detect(iCurrent, lObjects):
        iCurrent = timeout_clause.classify_until([';'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import rules
from vsg.rules import utils


class token_case_in_range_bounded_by_tokens_with_prefix_suffix(rules.token_case_with_prefix_suffix):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token types

    oStart : token type

    oEnd : token type
    '''

    def __init__(self, name, identifier, lTokens, oStart, oEnd):
        rules.token_case_with_prefix_suffix.__init__(self, name=name, identifier=identifier, lTokens=lTokens)
        self.oStart = oStart
        self.oEnd = oEnd

    def _get_tokens_of_interest(self, oFile):
        self.case_exceptions_lower = utils.lowercase_list(self.case_exceptions)
        return oFile.get_tokens_matching_in_range_bounded_by_tokens(self.lTokens, self.oStart, self.oEnd)





from vsg.vhdlFile.classify import full_type_declaration
from vsg.vhdlFile.classify import incomplete_type_declaration


def detect(iToken, lObjects):
    '''
    type_declaration ::=
        full_type_declaration
      | incomplete_type_declaration
    '''

    iReturn = full_type_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = incomplete_type_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    return iToken




from vsg.rules import split_line_at_token_when_between_tokens_unless_token_is_found

from vsg import token

lTokens = []
lTokens.append(token.assertion.severity_keyword)

oStart = token.assertion.keyword
oEnd = token.assertion_statement.semicolon
oStop = token.concurrent_assertion_statement.semicolon


class rule_005(split_line_at_token_when_between_tokens_unless_token_is_found):
    '''
    This rule checks the **severity** keyword is on its own line for sequential assertion statements.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin

         process begin

           assert WIDTH > 16 report "FIFO width is limited to 16 bits." severity FAILURE;

         end process;

       end architecture rtl;

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin

         process begin

           assert WIDTH > 16 report "FIFO width is limited to 16 bits."
             severity FAILURE;

         end process;

       end architecture rtl;
    '''

    def __init__(self):
        split_line_at_token_when_between_tokens_unless_token_is_found.__init__(self, 'assert', '005', lTokens, oStart, oEnd, oStop)
        self.solution = "Place **severity** keyword on its own line."




from vsg.rules import utils as rules_utils
from vsg.rules import create_violation


def add_new_line_and_remove_new_line(self, oToi, sOption, oTokenType):
    if sOption == 'ignore':
        return
    elif sOption == 'add_new_line':
        rules_utils.analyze_with_function(self, oToi, oTokenType, analyze_add_new_line)
    elif sOption == 'remove_new_line':
        rules_utils.analyze_with_function(self, oToi, oTokenType, analyze_remove_new_line)


def analyze_add_new_line(self, oToi):
    iToken = oToi.get_meta_data('iToken')
    lTokens = oToi.get_tokens()
    if not rules_utils.token_at_beginning_of_line_in_token_list(iToken, lTokens):
        oViolation = create_violation.add_new_line(oToi)
        self.add_violation(oViolation)


def analyze_remove_new_line(self, oToi):
    iToken = oToi.get_meta_data('iToken')
    lTokens = oToi.get_tokens()
    if rules_utils.token_at_beginning_of_line_in_token_list(iToken, lTokens):
        oViolation = create_violation.remove_new_line(self, oToi)
        self.add_violation(oViolation)




from vsg import parser
from vsg.rule_group import length
from vsg import violation


class file_length(length.Rule):
    '''
    Checks the length of the file.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    '''

    def __init__(self, name, identifier):
        length.Rule.__init__(self, name=name, identifier=identifier)
        self.length = 2000

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_matching([parser.carriage_return])

    def _analyze(self, lToi):
        if len(lToi) > self.length:
            sSolution = 'Reduce file length to less than ' + str(self.length) + ' lines'
            oViolation = violation.New(self.length, None, sSolution)
            self.add_violation(oViolation)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment

from vsg.token import port_map_aspect as token

lTokens = []
lTokens.append(token.open_parenthesis)


class rule_005(insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment):
    '''
    This rule checks for a port assignment on the same line as the **port map** keyword.

    **Violation**

    .. code-block:: vhdl

         port map (WR_EN    => wr_en,
           RD_EN    => rd_en,
           OVERFLOW => overflow
         );

    **Fix**

    .. code-block:: vhdl

         port map (
           WR_EN    => wr_en,
           RD_EN    => rd_en,
           OVERFLOW => overflow
         );
    '''

    def __init__(self):
        insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment.__init__(self, 'port_map', '005', lTokens)
        self.solution = 'Move port assignment to it\'s own line.'




from vsg.rules import blank_line_below_line_ending_with_several_possible_tokens

from vsg import token

lTokens = []
lTokens.append(token.block_statement.block_keyword)
lTokens.append(token.block_statement.guard_close_parenthesis)
lTokens.append(token.block_statement.is_keyword)

lAllowTokens = []
lAllowTokens.append(token.block_statement.begin_keyword)


class rule_201(blank_line_below_line_ending_with_several_possible_tokens):
    '''
    This rule checks for a blank line below the **block** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       block_label : block is
         constant width : integer := 32;

    **Fix**

    .. code-block:: vhdl

       block_label : block is

         constant width : integer := 32;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_several_possible_tokens.__init__(self, 'block', '201', lTokens, lAllowTokens)




from vsg.token import entity_name_list as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import entity_designator


def classify(iToken, lObjects):
    '''
    entity_name_list ::=
        entity_designator { , entity_designator }
      | others
      | all
    '''

    if utils.is_next_token('others', iToken, lObjects):

        return utils.assign_next_token_required('others', token.others_keyword, iToken, lObjects)

    elif utils.is_next_token('all', iToken, lObjects):

        return utils.assign_next_token_required('all', token.all_keyword, iToken, lObjects)

    else:

        iCurrent = entity_designator.classify(iToken, lObjects)

        while utils.is_next_token(',', iCurrent, lObjects):

            iCurrent = utils.assign_next_token_required(',', token.comma, iToken, lObjects)

            entity_designator.classify(iToken, lObjects)

    return iCurrent




from vsg import rules


class token_case_subtype_indication(rules.token_case):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token types
       The token types to check the case one.

    lEndTokens: list of token types
       The token types to stop searching.
    '''

    def __init__(self, name, identifier, lTokens, lEndTokens):
        rules.token_case.__init__(self, name=name, identifier=identifier, lTokens=lTokens)
        self.lEndTokens = lEndTokens

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_starting_with_token_and_ending_with_one_of_possible_tokens(self.lTokens, self.lEndTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.procedure_call.procedure_name)


class rule_302(token_indent):
    '''
    This rule checks the indent of the *procedure* name.

    **Violation**

    .. code-block:: vhdl

       a <= b;

         WR_EN(parameter);

    **Fix**

    .. code-block:: vhdl

       a <= b;

       WR_EN(parameter);
    '''

    def __init__(self):
        token_indent.__init__(self, 'procedure_call', '302', lTokens)




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.process_statement.process_label, token.process_statement.label_colon])


class rule_024(Rule):
    '''
    This rule checks for a single space after the process label.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       proc_a: process (rd_en, wr_en, data_in, data_out,
                        rd_full, wr_full
                       ) is
       begin

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin
    '''
    def __init__(self):
        Rule.__init__(self, 'process', '024', lTokens)




from vsg.vhdlFile.extract import utils

from vsg.vhdlFile.extract.get_line_succeeding_line import get_line_succeeding_line


def get_line_below_line_ending_with_token_with_hierarchy(lTokens, lAllTokens, lHierarchy, oTokenMap):

    lReturn = []

    lTokenIndexes = utils.get_indexes_of_token_list(lTokens, oTokenMap)

    lIndexes = []
    for iIndex in lTokenIndexes:
        if lAllTokens[iIndex].get_hierarchy() in lHierarchy:
            if utils.is_token_at_end_of_line(iIndex, oTokenMap):
                lIndexes.append(iIndex)

    lLine = utils.get_line_numbers_of_indexes_in_list(lIndexes, oTokenMap)

    for iLine in lLine:
        lReturn.append(get_line_succeeding_line(iLine, lAllTokens, 1, oTokenMap))

    return lReturn




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.concurrent_procedure_call_statement.label_name)
lTokens.append(token.procedure_call_statement.label)


class rule_500(token_case):
    '''
    This rule checks the label has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       PROCEDURE_CALL_LABEL : WR_EN(paremeter);

    **Fix**

    .. code-block:: vhdl

       procedure_call_label : WR_EN(paremeter);
    '''

    def __init__(self):
        token_case.__init__(self, 'procedure_call', '500', lTokens)
        self.groups.append('case::label')
